package Hawat::Module::Db;
use strict;
use warnings;

#*******************************************************************************
# DOCUMENTATION SECTION
#*******************************************************************************

=head1 NAME

Hawat::Module::Db - Short module description

=head1 SYNOPSIS

 use Hawat::Module::Db;

=head1 DESCRIPTION

Extensive module description

=head1 AUTHOR

 Jan Mach
 Cesnet, z.s.p.o
 jan.mach@cesnet.cz
 http://www.cesnet.cz

=head1 COPYRIGHT

This program is free software; you can redistribute
it and/or modify it under the same terms as Perl itself.

The full text of the license can be found in the
LICENSE file included with this module.

=head1 SEE ALSO

perl(1).

=head1 FUNCTION REFERENCE

=over 4

=cut

#*******************************************************************************
# LIBRARY LOADING SECTION
#*******************************************************************************

#-- Perl core modules ---------------------------------------------------------#
use Carp;
use Scalar::Util 'blessed';
use Time::HiRes qw (time);

#use Data::Dumper; #-+-> DEVEL ONLY <-+-#
#use Smart::Comments; #-+-> DEVEL ONLY <-+-#

#-- Perl CPAN modules ---------------------------------------------------------#

#-- Custom application modules ------------------------------------------------#
use Hawat::Module;
use Mentat::Message::Convertor::Mongo;
use Mentat::Storage::Factory;
use Data::Filter::Factory;
use Mentat::Message::Query::Mongo;

#*******************************************************************************
# CONSTANTS AND GLOBAL VARIABLES DEFINITION SECTION
#*******************************************************************************

#-- Constants -----------------------------------------------------------------#

# Maximum number of messages displayed
use constant MESSAGES_DISPLAY_LIMIT => 10;

#-- Static public class variables (our) ---------------------------------------#

#-- Static protected class variables (my) -------------------------------------#

# Hash containing command structure
my %COMMANDS = (
            'clear' => {
                    '_DESC_' => 'Clear items from Mentat database',
                    ''      => [
                            '_command_clear_all' ,
                            'Clear all items from Mentat database (same as "all", see it for more info)',
                        ],
                    'all'   => [
                            '_command_clear_all' ,
                            'Clear all items from Mentat database',
                        ],
                    'first' => [
                            '_command_clear_first',
                            'Clear first [number] of items from Mentat database',
                            '(INTEGER) number - number to items to clear from beginning [OPTIONAL]',
                        ],
                    'last'  => [
                            '_command_clear_last',
                            'Clear last [number] of items from Mentat database',
                            '(INTEGER) number - number to items to clear from end [OPTIONAL]',
                        ],
                },
            'find'  => [
                    '_command_find' ,
                    'Find all items in Mentat database according to the given filter condition',
                    '(STRING) filter - filter condition [MANDATORY]',
                ],
            'index' => [
                    '_command_index' ,
                    'Create database indexes',
                ],
            'status'    => [
                    '_command_status' ,
                    'Check the status of the Mentat persistent database storage',
                ],
            'view'  => {
                    '_DESC_' => 'View items in Mentat database',
                    ''      => [
                            '_command_view_all' ,
                            'View last count items in Mentat database (same as "all", see it for more info)',
                        ],
                    'all'   => [
                            '_command_view_all' ,
                            'View all items in Mentat database',
                            '(INTEGER) limit - number to items to display [OPTIONAL,DEFAULT=' . MESSAGES_DISPLAY_LIMIT . ']',
                            '(INTEGER) skip - number to items to skip from beginning [OPTIONAL,DEFAULT=0]',
                        ],
                    'count' => [
                            '_command_view_count',
                            'Count the number of items in Mentat persistent database storage',
                        ],
                    'first' => [
                            '_command_view_first',
                            'View first [number] of items in Mentat database',
                            '(INTEGER) number - number to items to view from beginning [OPTIONAL]',
                            '(INTEGER) limit - number to items to display [OPTIONAL,DEFAULT=' . MESSAGES_DISPLAY_LIMIT . ']',
                            '(INTEGER) skip - number to items to skip from beginning [OPTIONAL,DEFAULT=0]',
                        ],
                    'last'  => [
                            '_command_view_last',
                            'View last [number] of items in Mentat database',
                            '(INTEGER) number - number to items to view from end [OPTIONAL]',
                            '(INTEGER) limit - number to items to display [OPTIONAL,DEFAULT=' . MESSAGES_DISPLAY_LIMIT . ']',
                            '(INTEGER) skip - number to items to skip from beginning [OPTIONAL,DEFAULT=0]',
                        ],
                },
        );

#*******************************************************************************
# INITIALIZATION AND CLEANUP SECTION
#*******************************************************************************

#-- Module initializations ----------------------------------------------------#
BEGIN {
    use vars qw($VERSION $DEVEL @ISA);
    $VERSION = '0.1';
    $DEVEL   = 0;
    @ISA     = ('Hawat::Module');
}


#-- Module clean-up code (global destructor) ----------------------------------#
END {

}

#*******************************************************************************
# IMPLEMENTATION SECTION
#*******************************************************************************

=item get_commands() [PUBLIC]

  Usage     :
  Purpose   : Return list of all commands this module provides
  Arguments :
  Returns   :
  Throws    : Croaks, if invoked on class
  Comments  :
  See Also  :

=cut

sub get_commands()
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    return %COMMANDS;
}

#-------------------------------------------------------------------------------
# Private interface
#-------------------------------------------------------------------------------

# _init [PROTECTED]
#
#  Usage     : return $self->_init(@_);
#  Purpose   : Initialize the new Hawat::Module::Db instance
#  Arguments :
#  Returns   :
#  Throws    : Croaks, if invoked on class
#  Comments  : This method must never be called directly, it is supposed to be overloaded by subclasses.
#  See Also  :

sub _init
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($config,) = @_;

    $self->{STORAGE}        = undef;
    $self->{CONVERTOR}      = undef;
    $self->{QUERY}          = undef;
    $self->{FILTER_FACTORY} = undef;

    return $self;
}

# _storage [PROTECTED]
#
#  Usage     :
#  Purpose   :
#  Arguments :
#  Returns   :
#  Throws    :
#  Comments  :
#  See Also  :

sub _storage
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);

    unless ($self->{STORAGE}) {
        $self->{STORAGE} = Mentat::Storage::Factory->storage(split(/[:,;]/, $self->cfg->{'database'}));
    }
    return $self->{STORAGE};
}

# _convertor [PROTECTED]
#
#  Usage     :
#  Purpose   :
#  Arguments :
#  Returns   :
#  Throws    :
#  Comments  :
#  See Also  :

sub _convertor
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);

    unless ($self->{CONVERTOR}) {
        $self->{CONVERTOR} = Mentat::Message::Convertor::Mongo->new();
    }
    return $self->{CONVERTOR};
}

# _query [PROTECTED]
#
#  Usage     :
#  Purpose   :
#  Arguments :
#  Returns   :
#  Throws    :
#  Comments  :
#  See Also  :

sub _query
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);

    unless ($self->{QUERY}) {
        $self->{QUERY} = Mentat::Message::Query::Mongo->new(convertor => $self->_convertor(),
                                                            storage => $self->_storage());
    }
    return $self->{QUERY};
}

# _ffactory [PROTECTED]
#
#  Usage     :
#  Purpose   :
#  Arguments :
#  Returns   :
#  Throws    :
#  Comments  :
#  See Also  :

sub _ffactory
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);

    unless ($self->{FILTER_FACTORY}) {
        $self->{FILTER_FACTORY} = Data::Filter::Factory->new_from_file($self->cfg->{'filter_grammar_file'});
    }
    return $self->{FILTER_FACTORY};
}

#-------------------------------------------------------------------------------
# Command callbacks
#-------------------------------------------------------------------------------

sub _command_clear_all
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($chain, @args) = @_;
    die "Invalid usage: [$0 @$chain] command does not accept any arguments\n" if @args;

    my $start = time();

    $self->out(" - Removing all messages from Mentat persistent storage");
    eval {
        my $rv = $self->_storage()->remove();
        my $count = $rv->{'n'};

        $self->out(" [SUCCESS: $count]\n", 'bold green');
        $self->res->{'count'} = $count;
    };
    if ($@) {
        $@ =~ s/at [^\s]+ line [\d]+\.\s*//;
        $self->out(" [FAILED: $@]\n", 'bold red');
        $self->res->{'error'} = $@;
    }
    my $stop = time();
    $self->out(" - Clearing duration: " . ($stop - $start) . " [s]\n");
    $self->res->{'duration'} = $stop - $start;
}

sub _command_clear_first
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);

    $self->out("FEATURE NOT IMPLEMENTED\n", 'bold red');
}

sub _command_clear_last
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);

    $self->out("FEATURE NOT IMPLEMENTED\n", 'bold red');
}

#-------------------------------------------------------------------------------

sub _command_find
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($chain, @args) = @_;
    my ($filter, $limit, $skip) = @args;
    die "Invalid usage: [$0 @$chain] must receive filter string as first argument\n" unless $filter;

    my $start = time();

    $limit  = MESSAGES_DISPLAY_LIMIT unless ($limit and (int($limit) > 0));
    $skip   = ($skip) ? int($skip) : 0;

    my $min = $skip + 1;
    my $max = $skip + $limit;

    $self->out(" - Searching for messages in Mentat persistent storage with filter '$filter', limit '$limit' and skip '$skip' (\#$min:\#$max)");
    eval {
        my ($messages, $count_all, $filter_tree);

        if ($filter =~ /^{/) {
            ($messages, $count_all) = $self->_query->query_msg_json($filter, undef, $limit, $skip);
        }
        else {
            $filter_tree            = $self->_ffactory()->filter_from_string($filter);
            ($messages, $count_all) = $self->_query->query_msg($filter_tree, undef, $limit, $skip);
        }

        my $count = scalar(@$messages);
        my $index = $min;

        $self->out(" [SUCCESS: $count_all]\n", 'bold green');
        map { $self->out("\n ----------\n   " . ($index++) . "\n ----------\n\n" . $_->to_string(1) . "\n"); } @$messages;
        $self->res->{'filter'}  = $filter;
        $self->res->{'limit'}       = $limit;
        $self->res->{'skip'}        = $skip;
        $self->res->{'min'}     = $min;
        $self->res->{'max'}     = $max;
        $self->res->{'count_all'}   = $count_all;
        $self->res->{'count'}   = $count;
        $self->res->{'messages'}    = $messages;
    };
    if ($@) {
        $@ =~ s/at [^\s]+ line [\d]+\.\s*//;
        $self->out(" [FAILED: $@]\n", 'bold red');
        $self->res->{'error'} = $@;
    }
    my $stop = time();
    $self->out(" - Viewing duration: " . ($stop - $start) . " [s]\n");
    $self->res->{'duration'} = $stop - $start;

}

#-------------------------------------------------------------------------------

sub _command_index
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($chain, @args) = @_;
    die "Invalid usage: [$0 @$chain] command does not accept any arguments\n" if @args;

    my $start = time();

    $self->out(" - Creating indexes in persistent storage");
    eval {
        $self->_storage->create_indexes();
        $self->out(" [SUCCESS]\n", 'bold green');
        $self->res->{'indexes'} = 1;
    };
    if ($@) {
        $@ =~ s/at [^\s]+ line [\d]+\.\s*//;
        $self->out(" [FAILED: $@]\n", 'bold red');
        $self->res->{'error'} = $@;
        $self->res->{'indexes'} = 0;
    }
    my $stop = time();
    $self->out(" - Persistent storage index creation duration: " . ($stop - $start) . " [s]\n");
    $self->res->{'duration'} = $stop - $start;
}

#-------------------------------------------------------------------------------

sub _command_status
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($chain, @args) = @_;
    die "Invalid usage: [$0 @$chain] command does not accept any arguments\n" if @args;

    my $start = time();

    $self->out(" - Checking persistent storage connection");
    eval {
        $self->_storage();
        $self->out(" [SUCCESS]\n", 'bold green');
        $self->res->{'connection'} = 1;
    };
    if ($@) {
        $@ =~ s/at [^\s]+ line [\d]+\.\s*//;
        $self->out(" [FAILED: $@]\n", 'bold red');
        $self->res->{'error'} = $@;
        $self->res->{'connection'} = 0;
    }
    my $stop = time();
    $self->out(" - Persistent storage connection duration: " . ($stop - $start) . " [s]\n");
    $self->res->{'duration'} = $stop - $start;
}

#-------------------------------------------------------------------------------

sub _command_view_all
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($chain, @args) = @_;
    my ($limit, $skip, $sort) = @args;

    my $start = time();

    $limit = MESSAGES_DISPLAY_LIMIT unless ($limit and (int($limit) > 0));
    $skip  = ($skip) ? int($skip) : 0;
    $sort  = ($sort) ? $sort : {'ts' => 1};

    my $min = $skip + 1;
    my $max = $skip + $limit;

    $self->out(" - Vieving messages in Mentat persistent storage with limit '$limit' and skip '$skip' (\#$min:\#$max)");
    eval {
        my ($messages, $count_all) = $self->_storage()->find_msg(undef, $sort, $limit, $skip);
        my $index = $min;

        $self->out(" [SUCCESS: $count_all]\n", 'bold green');
        map { $self->out("\n ----------\n   " . ($index++) . "\n ----------\n\n" . $_->to_string(1) . "\n"); } @$messages;
        $self->res->{'limit'}     = $limit;
        $self->res->{'skip'}      = $skip;
        $self->res->{'min'}       = $min;
        $self->res->{'max'}       = $max;
        $self->res->{'count_all'} = $count_all;
        $self->res->{'messages'}  = $messages;
    };
    if ($@) {
        $@ =~ s/at [^\s]+ line [\d]+\.\s*//;
        $self->out(" [FAILED: $@]\n", 'bold red');
        $self->res->{'error'} = $@;
    }
    my $stop = time();
    $self->out(" - Viewing duration: " . ($stop - $start) . " [s]\n");
    $self->res->{'duration'} = $stop - $start;
}

sub _command_view_count
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($chain, @args) = @_;
    die "Invalid usage: [$0 @$chain] command does not accept any arguments\n" if @args;

    my $start = time();

    $self->out(" - Counting messages in Mentat persistent storage");
    eval {
        my $count = $self->_storage()->count();

        $self->out(" [SUCCESS: $count]\n", 'bold green');
        $self->res->{'count'} = $count;
    };
    if ($@) {
        $@ =~ s/at [^\s]+ line [\d]+\.\s*//;
        $self->out(" [FAILED: $@]\n", 'bold red');
        $self->res->{'error'} = $@;
    }
    my $stop = time();
    $self->out(" - Counting duration: " . ($stop - $start) . " [s]\n");
    $self->res->{'duration'} = $stop - $start;
}

sub _command_view_first
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($chain, @args) = @_;
    my ($limit, $skip) = @args;

    return $self->_command_view_all($chain, $limit, $skip);
}

sub _command_view_last
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($chain, @args) = @_;
    my ($limit, $skip) = @args;

    return $self->_command_view_all($chain, $limit, $skip, {'ts' => -1});
}

=pod

=back

=cut

1;
