package Mentat::Message::Convertor::Mongo;
use strict;
use warnings;

#*******************************************************************************
# DOCUMENTATION SECTION
#*******************************************************************************

=head1 NAME

Mentat::Message::Convertor::Mongo - Convert messages od various types into MongoDB optimal representation

=head1 SYNOPSIS

 use Mentat::Message::Convertor::Mongo;

=head1 DESCRIPTION



=head1 AUTHOR

 Jan Mach
 Cesnet, z.s.p.o
 jan.mach@cesnet.cz
 http://www.cesnet.cz

=head1 COPYRIGHT

This program is free software; you can redistribute
it and/or modify it under the same terms as Perl itself.

The full text of the license can be found in the
LICENSE file included with this module.

=head1 SEE ALSO

perl(1), Data::Convertor(3).

=head1 FUNCTION REFERENCE

=over 4

=cut

#*******************************************************************************
# LIBRARY LOADING SECTION
#*******************************************************************************

#-- Perl core modules ---------------------------------------------------------#
use Carp;
use Scalar::Util 'blessed';
use NetAddr::IP;

use Data::Dumper; #-+-> DEVEL ONLY <-+-#
#use Smart::Comments; #-+-> DEVEL ONLY <-+-#

#-- Perl CPAN modules ---------------------------------------------------------#
use MongoDB::BSON::Binary;

#-- Custom application modules ------------------------------------------------#
use Data::Convertor;
use Data::Tree;

# Modules from Cesnet Toolkit
use Log::Loger;

#*******************************************************************************
# CONSTANTS AND GLOBAL VARIABLES DEFINITION SECTION
#*******************************************************************************

#-- Constants -----------------------------------------------------------------#

#"\x00"  Binary / Generic
#"\x01"  Function
#"\x02"  Binary (Old)
#"\x03"  UUID
#"\x05"  MD5
#"\x80"  User defined
use constant SUBTYPE_BINARY => 0;

use constant IP   => 'ip';
use constant MIN  => 'min';
use constant MAX  => 'max';
use constant TYPE => 'type';

use constant IPV4 => 'IPV4';
use constant IPV6 => 'IPV6';

#-------------------------------------------------------------------------------
# Structures for IDEA -> MongoDB HASH conversions
#-------------------------------------------------------------------------------

use constant IDEA_NODES => {
    # src_mpath => [conversion, dst_mpath]

    'CreateTime'   => ['_convert_idea_datetime', 'CreateTime'],
    'DetectTime'   => ['_convert_idea_datetime', 'DetectTime'],
    'EventTime'    => ['_convert_idea_datetime', 'EventTime'],
    'CeaseTime'    => ['_convert_idea_datetime', 'CeaseTime'],
    'WinStartTime' => ['_convert_idea_datetime', 'WinStartTime'],
    'WinEndTime'   => ['_convert_idea_datetime', 'WinEndTime'],

    'Source/IP4' => ['_convert_idea_ip', IPV4()],
    'Source/IP6' => ['_convert_idea_ip', IPV6()],
    'Target/IP4' => ['_convert_idea_ip', IPV4()],
    'Target/IP6' => ['_convert_idea_ip', IPV6()],

    'Source/Port' => ['anything_to_int'],
    'Target/Port' => ['anything_to_int'],
    'Source/ASN'  => ['anything_to_int'],
    'Target/ASN'  => ['anything_to_int'],
    'Attach/Size' => ['anything_to_int'],
    'ConnCount'   => ['anything_to_int'],
    'FlowCount'   => ['anything_to_int'],
    'PacketCount' => ['anything_to_int'],
    'ByteCount'   => ['anything_to_int'],
};

use constant IDEA_ADDRESS_CONVERSIONS => {
    IPV4() => '_ipv4_to_hob',
    IPV6() => '_ipv6_to_hob',
};

#-- Static public class variables (our) ---------------------------------------#

#-- Static protected class variables (my) -------------------------------------#

#*******************************************************************************
# INITIALIZATION AND CLEANUP SECTION
#*******************************************************************************

#-- Module initializations ----------------------------------------------------#
BEGIN {
    use vars qw($VERSION $DEVEL @ISA);
    $VERSION = '0.1';
    $DEVEL   = 0;
    @ISA     = ('Data::Convertor');
}


#-- Module clean-up code (global destructor) ----------------------------------#
END {

}

#*******************************************************************************
# IMPLEMENTATION SECTION
#*******************************************************************************

=item convert_node_idea($$$) [PUBLIC]

  Usage     : my $result = $instance->convert_node_idea($idea_message, )
  Purpose   : Convert whole Mentat::Message::IDEA to MongoDB optimal HASH
  Arguments : Mentat::Message::IDEA $idea_message - message to be converted
  Returns   : REFERENCE containing the converted data structure
  Throws    : Croaks, if not invoked on object

=cut

sub convert_node_idea($$$)
{
    my $self = shift;
    confess "Instance method invoked on class" unless blessed($self);
    my ($idea, $src_mpath, $value) = @_;
    confess "Mentat::Message::IDEA object must be given as first argument" unless blessed($idea) and $idea->isa('Mentat::Message::IDEA');

    # The are no indexes in the path conversion specifications, so remove them first
    my $src_mpath_s = $src_mpath; $src_mpath_s =~ s/\[\d*\]//g;

    # Get the indexes
    my @indexes = ($src_mpath =~ /\[(\d+)\]/g);
    my $variables = {}; my $i = 0;
    map { $variables->{++$i} = $_; } @indexes;

    # Handle the MPath unknown to the convertor
    if (not exists IDEA_NODES->{$src_mpath_s}) {
        #$src_mpath =~ s/^Alert\/|^Heartbeat\///;
        return ([$src_mpath, $value]);
    }
    # Skip the given node
    if(not IDEA_NODES()->{$src_mpath_s}) {
        print STDERR "Disabled path '$src_mpath_s', skipping\n" if $DEVEL;
        return ();
    }

    ### Handle the MPaths known to the convertor
    # Fetch the conversion information
    my ($conversion, @args) = @{IDEA_NODES()->{$src_mpath_s}};
    $conversion = $conversion || 'anything_to_samething';

    # Dispatching for conversions performed directly via Value::Convertor module
    unless ($self->can($conversion)) {
        return $self->_convert_node($src_mpath, $value, $conversion, $variables, @args);
    }
    else {
        return $self->$conversion($idea, $src_mpath, $value, $variables, @args);
    }
}

=item convert_value($$$) [PUBLIC]

  Usage     : my $result = $instance->convert_value($mpath, $value, $operator)
  Purpose   : Convert given value to the datatype expected for node on given MPath
  Arguments : STRING $mpath - path to value
              STRING $value - value as string
              STRING $operator - operator as string
  Returns   : Value in hash and optimal representation (INT, STRING, ...)
  Throws    : Croaks, if not invoked on object
  Comments  : Some conversions require

=cut

sub convert_value($$$)
{
    my $self = shift;
    confess "Instance method invoked on class" unless blessed($self);
    my ($mpath, $value, $operator) = @_;
    confess "MPath must be given as argument" unless defined($mpath);
    confess "Value must be given as argument" unless defined($value);

    my ($conversion, $type, $cmin, $cmax, $sub, $ipv);

    # Prepare the MPath
    $conversion->{MPATH} = $mpath;
    # We must translate '/' in MPath to '.', which is the separator in MongoDB queries
    $conversion->{MPATH} =~ s|/|.|g;
    # And we must remove the index specifications
    $conversion->{MPATH} =~ s/(\[[\d]+\])//g;

    # Prepare the rest of the arguments
    $conversion->{PATH}     = $conversion->{MPATH};
    $conversion->{VALUE}    = $value;
    $conversion->{OPERATOR} = $operator;
    $conversion->{REPLACE}  = 0;

    if ($conversion->{MPATH} =~ /(Source|Target)\.(IP4|IP6)/) {
        $sub = $1;

        ($type, $cmin, $cmax) = Value::Convertor->ip_to_bin($conversion->{VALUE});
        $ipv = ($type =~ /IPV4/)?'IP4':'IP6';

        # Single IP addresses return only one value from conversion
        unless ($cmax) {
            $conversion->{PATH} = "$sub.$ipv.ip";
            $conversion->{VALUE} = MongoDB::BSON::Binary->new(data => $cmin, subtype => 0);
        }
        # Fix of the previous bug, that returned gibberish for events with multiple IPs
        # https://docs.mongodb.com/manual/reference/operator/query/elemMatch/#array-of-embedded-documents
        #{
        #   "Source.IP4": {
        #       $elemMatch: {
        #           "min": {
        #               "$gte": BinData(...)
        #           },
        #           "max": {
        #               "$lte": BinData(...)
        #           }
        #       }
        #   }
        else {
            # Original version (returned gibberish)
            #$conversion->{PATH} = "$sub.$ipv";
            #$conversion->{VALUE} = {
            #    $conversion->{PATH}.".min" => { '$gte' => MongoDB::BSON::Binary->new(data => $cmin, subtype => 0) },
            #    $conversion->{PATH}.".max" => { '$lte' => MongoDB::BSON::Binary->new(data => $cmax, subtype => 0) },
            #};
            $conversion->{PATH} = "$sub.$ipv";
            $conversion->{VALUE} = {
                # This is a correct query version
                $conversion->{PATH} => {
                    '$elemMatch' => {
                        'min' => {
                            '$gte' => MongoDB::BSON::Binary->new(data => $cmin, subtype => 0)
                        },
                        'max' => {
                            '$lte' => MongoDB::BSON::Binary->new(data => $cmax, subtype => 0)
                        }
                    }
                }
            };
            $conversion->{REPLACE} = 1;
        }
    }
    elsif ($conversion->{MPATH} =~ /(DetectTime|CreateTime)/) {
        ($type, $cmin) = $self->_convert_with_conversion('datetime_to_nbi', $conversion->{VALUE});
        $cmin = Value::Convertor->bigint_to_bin($cmin);
        $conversion->{VALUE} = MongoDB::BSON::Binary->new(data => $cmin, subtype => 0);
    }
    elsif ($conversion->{MPATH} =~ /(StorageTime)/) {
        $conversion->{VALUE} = int($conversion->{VALUE});
    }
    elsif ($conversion->{MPATH} eq "ts") {
        ($type, $cmin) = $self->_convert_with_conversion('datetime_to_nbi', $conversion->{VALUE});
        $cmin = Value::Convertor->bigint_to_bin($cmin);
        $conversion->{VALUE} = MongoDB::BSON::Binary->new(data => $cmin, subtype => 0);
    }

    #Log::Loger->info(__PACKAGE__, "Conversion: ", Dumper ($conversion) ) if $DEVEL;
    #Log::Loger->info(__PACKAGE__, "PATH: ", $conversion->{PATH} ) if $DEVEL;
    #Log::Loger->info(__PACKAGE__, "VALUE: ", $conversion->{VALUE} ) if $DEVEL;

    return $conversion;
}

=item convert_ts($) [PUBLIC]

  Usage     : my $result = $instance->convert_ts($ts)
  Purpose   : Convert given unix timestamp to appropriate database representation
  Arguments : INTEGER $ts - unix timestamp
  Returns   : MongoDB::BSON::Binary $timestamp
  Throws    : Croaks, if not invoked on object

=cut

sub convert_ts($)
{
    my $self = shift;
    confess "Instance method invoked on class" unless blessed($self);
    my ($ts,) = @_;
    confess "Unix timestamp is expected to be given as argument" unless defined $ts;

    $ts = Value::Convertor->unixs_to_nbi($ts);
    $ts = Value::Convertor->bigint_to_bin($ts);

    return MongoDB::BSON::Binary->new(data => $ts, subtype => SUBTYPE_BINARY());
}

#-------------------------------------------------------------------------------
# Private interface
#-------------------------------------------------------------------------------

# _init [PROTECTED]
#
#  Usage     : return $self->_init(@_);
#  Purpose   : Initialize the new Mentat::Message::Convertor instance
#  Arguments :
#  Returns   : Mentat::Message::Convertor reference
#  Throws    : Croaks, if invoked on class
#  Comments  : This method must never be called directly, it is supposed to be overloaded by subclasses.
#  See Also  :

sub _init
{
    my $self = shift;
    confess "Instance method invoked on class" unless blessed($self);

    return $self;
}

#-------------------------------------------------------------------------------

# array of binary
sub _ipv6_to_hob($)
{
    my $self = shift;
    confess "Instance method invoked on class" unless blessed($self);
    my ($ipv6,) = @_;

    my $ipv6obj = NetAddr::IP->new("$ipv6");
    my $net     = $ipv6obj->network();
    my $bcast   = $ipv6obj->broadcast();

    #$ipaddr =~ s/\/\d+//;
    #$net =~ s/\/\d+//;
    #$broadcast =~ s/\/\d+//;

    my $ipv6bin = Value::Convertor->ipv6str_to_bin("$ipv6");
    my $min     = Value::Convertor->ipv6str_to_bin($net);
    my $max     = Value::Convertor->ipv6str_to_bin($bcast);

    return {
        IP()    => MongoDB::BSON::Binary->new(data => $ipv6bin, subtype => SUBTYPE_BINARY()),
        MIN()   => MongoDB::BSON::Binary->new(data => $min, subtype => SUBTYPE_BINARY()),
        MAX()   => MongoDB::BSON::Binary->new(data => $max, subtype => SUBTYPE_BINARY()),
    };
}

sub _ipv4_to_hob($)
{
    my $self = shift;
    confess "Instance method invoked on class" unless blessed($self);
    my ($ipv4,) = @_;

    # http://backpan.perl.org/authors/id/S/SW/SWETH/IPv4.pm
    # http://code.google.com/p/ipv6gen/

    my ($type, @result) = Value::Convertor->ipv4_to_bin($ipv4);

    if ($type ne Value::Convertor::UNKNOWN_TYPE()) {
        if (@result == 1) {
            return {
                IP()    => MongoDB::BSON::Binary->new(data => $result[0], subtype => SUBTYPE_BINARY()),
                MIN()   => MongoDB::BSON::Binary->new(data => $result[0], subtype => SUBTYPE_BINARY()),
                MAX()   => MongoDB::BSON::Binary->new(data => $result[0], subtype => SUBTYPE_BINARY()),
            };
        }
        else {
            return {
                MIN()   => MongoDB::BSON::Binary->new(data => $result[0], subtype => SUBTYPE_BINARY()),
                MAX()   => MongoDB::BSON::Binary->new(data => $result[1], subtype => SUBTYPE_BINARY()),
            };
        }
    }
    else {
        die "Invalid IPV4 value '$ipv4'\n";
    }
}

#-------------------------------------------------------------------------------

# _convert_idea_datetime($) [PROTECTED]
#
#  Usage     :
#  Purpose   :
#  Arguments :
#  Returns   :
#  Throws    : Croaks, if invoked on class
#  Comments  :
#  See Also  :

sub _convert_idea_datetime($)
{
    my $self = shift;
    confess "Instance method invoked on class" unless blessed($self);
    my ($idea, $src_mpath, $value, $variables, $type, @args) = @_;
    confess "DATETIME string must be given as argument" unless defined $value;

    my $bin = Value::Convertor->bigint_to_bin(Value::Convertor->datestr_to_nbi("$value"));

    return (["$type", MongoDB::BSON::Binary->new(data => $bin, subtype => SUBTYPE_BINARY())]);
}

# _convert_idea_ip($$$) [PROTECTED]
#
#  Usage     :
#  Purpose   :
#  Arguments :
#  Returns   :
#  Throws    : Croaks, if invoked on class
#  Comments  :
#  See Also  :

sub _convert_idea_ip($$$)
{
    my $self = shift;
    confess "Instance method invoked on class" unless blessed($self);
    my ($idea, $src_mpath, $value, $variables, $type, @args) = @_;
    #confess "Address must be given as argument" unless defined $value;
    #confess "Type must be given as argument" unless defined $type;
    #confess "Invalid address type '$type'" unless IDEA_ADDRESS_CONVERSIONS()->{$type};

    my $conversion = IDEA_ADDRESS_CONVERSIONS()->{$type};
    my $result = $self->_convert_with_conversion($conversion, $value, @args);
    my $dst_mpath = $src_mpath;

    return ([$dst_mpath, $result]);
}

=pod

=back

=cut

1;
