package Mentat::Message::Correlator::Context::Queue::Memory;
use strict;
use warnings;

#*******************************************************************************
# DOCUMENTATION SECTION
#*******************************************************************************

=head1 NAME

Mentat::Message::Correlator::Context::Queue::Memory - Queue implementation for Mentat::Processor::Messages

=head1 SYNOPSIS

 use Mentat::Message::Correlator::Context::Queue::Memory;

=head1 DESCRIPTION

Extensive module description

=head1 AUTHOR

Jan Mach
Cesnet, z.s.p.o
jan.mach@cesnet.cz
http://www.cesnet.cz

=head1 COPYRIGHT

This program is free software; you can redistribute
it and/or modify it under the same terms as Perl itself.

The full text of the license can be found in the
LICENSE file included with this module.

=head1 SEE ALSO

perl(1).

=head1 FUNCTION REFERENCE

=over 4

=cut

#*******************************************************************************
# LIBRARY LOADING SECTION
#*******************************************************************************

#-- Perl core modules ---------------------------------------------------------#
use Carp;
use Scalar::Util 'blessed';

#use Data::Dumper; #-+-> DEVEL ONLY <-+-#
#use Smart::Comments; #-+-> DEVEL ONLY <-+-#

#-- Perl CPAN modules ---------------------------------------------------------#
use Tie::Array::Sorted;
use Tie::Array::Sorted::Lazy;

#-- Custom application modules ------------------------------------------------#
use Mentat::Core qw(:c_general);
use Mentat::Message::Correlator::Context::Queue;

use Tie::Array::Sorted::Stable;
use Tie::Array::Sorted::Stable::Lazy;

#*******************************************************************************
# CONSTANTS AND GLOBAL VARIABLES DEFINITION SECTION
#*******************************************************************************

#-- Constants -----------------------------------------------------------------#

#-- Static public class variables (our) ---------------------------------------#

#-- Static protected class variables (my) -------------------------------------#

#*******************************************************************************
# INITIALIZATION AND CLEANUP SECTION
#*******************************************************************************

#-- Module initializations ----------------------------------------------------#
BEGIN {
    use vars qw($VERSION $DEVEL @ISA);
    $VERSION = '0.1';
    $DEVEL   = 0;
    @ISA     = ('Mentat::Message::Correlator::Context::Queue')
}

#-- Module clean-up code (global destructor) ----------------------------------#
END {

}

#*******************************************************************************
# IMPLEMENTATION SECTION
#*******************************************************************************

=item DESTROY() [PUBLIC]

  Usage     : This method must not be used directly
  Purpose   : Class destructor
  Arguments : None
  Returns   : Nothing
  Throws    : Dies, if not invoked on instance

=cut

sub DESTROY()
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    ### Destroying Mentat.Message.Correlator.Context.Queue.Memory instance...

    # Queue of pending messsages
    delete $self->{QUEUE};
    # Map MESSAGE_ID => MESSAGE
    delete $self->{QUEUE_MAP};

    # Call parent version
    $self = $self->SUPER::DESTROY();
}

#-------------------------------------------------------------------------------

=item enqueue($) [PUBLIC]

  Usage     : $instance->enqueue($context);
  Purpose   : Enqueue given context into context queue
  Arguments : STRING $context_id - context identifier [MANDATORY]
              Mentat::Message::Correlator::Context $context - context to be queued [MANDATORY]
  Returns   : Nothing
  Throws    : Dies, if not invoked on object instance or if given invalid arguments

=cut

sub enqueue($)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($context,) = @_;
    confess "Context must be given as first argument" unless $context;

    # Retrieve the context ID
    my $context_id = $context->id();
    confess "Context must have unique identifier already assigned" unless $context_id;
    confess "Context with ID '$context_id' is already in queue" if exists $self->{QUEUE_MAP}->{$context_id};

    # Enqueue the reference to the context, not context itself, otherwise context
    # fetching would not work
    push(@{$self->{QUEUE}}, \$context);
    # Record the message to the message map (for backreference during message commit)
    # and again use reference to the message
    $self->{QUEUE_MAP}->{$context_id} = \$context;

    $self->{STATISTICS}->{enqueued}++;
}

=item dequeue() [PUBLIC]

  Usage     : my $context = $instance->dequeue();
  Purpose   : Fetch next timeouted context from the queue
  Arguments : None
  Returns   : Mentat::Message::Correlator::Context $context or UNDEF
  Throws    : Dies, if not invoked on object instance or if given invalid arguments

=cut

sub dequeue()
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);

    my $time = time();
    my ($next_time, $record);

    ### Check the first message in message queue
    while ($self->{QUEUE}->[0]) {
        ### Check, that it is not an empty record
        if (${$self->{QUEUE}->[0]}->{ID}) {
            ### Check, that it is time to process the first record
            $next_time = ${$self->{QUEUE}->[0]}->{TIMEOUT};
            if ($time >= $next_time) {
                $record = ${shift(@{$self->{QUEUE}})};
                delete $self->{QUEUE_MAP}->{$record->{ID}};
                $self->{STATISTICS}->{dequeued}++;
                return $record;
            }
            ### No luck
            last;
        }
        ### It was empty record, skip and try next one
        $self->{STATISTICS}->{skipped}++;
        shift(@{$self->{QUEUE}});
        next;
    }

    ### No record to process at this time
    return undef;
}

=item update($) [PUBLIC]

  Usage     : my $error = $instance->update($context_id, $data);
  Purpose   : Update context with given identifier with given data
  Arguments : STRING $context_id - context identifier [MANDATORY]
  Returns   : BOOL $result
  Throws    : Dies, if not invoked on object instance or if given invalid arguments
  Comment   : It is not allowed to update context ID or timeout

=cut

sub update($$)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($context_id, $data) = @_;
    confess "Context ID must be given as second argument" unless $context_id;

    # Check if the message with given identifier exists
    return undef unless exists $self->{QUEUE_MAP}->{$context_id};

    # Update context record
    my $context = ${$self->{QUEUE_MAP}->{$context_id}};
    # TODO

    $self->{STATISTICS}->{updated}++;
    return TRUE;
}

=item fetch($) [PUBLIC]

  Usage     : my $error = $instance->fetch($context_id);
  Purpose   : Fetch context with given identifier
  Arguments : STRING $context_id - context identifier [MANDATORY]
  Returns   : Mentat::Message::Correlator::Context $context or UNDEF
  Throws    : Dies, if not invoked on object instance or if given invalid arguments

=cut

sub fetch($)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($context_id,) = @_;
    confess "Context ID must be given as second argument" unless $context_id;

    # Check if the message with given identifier exists
    return undef unless exists $self->{QUEUE_MAP}->{$context_id};

    # Replace context record with empty record (we don`t know, on what position
    # it is, so we are using references to references). We use empty
    # record because it is valid record, so it does not mess up the auto sorting
    # algorithm in Tie::Array::Sorted* array.
    my $result = ${$self->{QUEUE_MAP}->{$context_id}};
    ${$self->{QUEUE_MAP}->{$context_id}} = {TIMEOUT => $result->{TIMEOUT}};

    # Delete the event from the event map
    delete $self->{QUEUE_MAP}->{$context_id};

    $self->{STATISTICS}->{fetched}++;
    return $result;
}

#-------------------------------------------------------------------------------

=item count(;$) [PUBLIC]

  Usage     : my $count = $instance->count();
  Purpose   : Get number of contexts in queue
  Arguments : None
  Returns   : INTEGER $count
  Throws    : Dies, if not invoked on object instance or if given invalid arguments

=cut

sub count(;$)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($real,) = @_;

    # By default count all context queue records
    # (including empty records)
    unless ($real) { return scalar @{$self->{QUEUE}}; }
    else           { return scalar keys %{$self->{QUEUE_MAP}}; }
}

=item exists($) [PUBLIC]

  Usage     : my $result = $instance->exists($context_id);
  Purpose   : Check, if the context with given ID is already present in the queue
  Arguments : STRING $context_id - context identifier [MANDATORY]
  Returns   : BOOL $result
  Throws    : Dies, if not invoked on object instance or if given invalid arguments

=cut

sub exists($)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($context_id,) = @_;
    return exists $self->{QUEUE_MAP}->{$context_id};
}

#-------------------------------------------------------------------------------
# Private interface
#-------------------------------------------------------------------------------

# _init [PROTECTED]
#
#  Usage     : return $self->_init(@_);
#  Purpose   : Initialize the new Mentat::Message::Correlator::Context::Queue::Memory; instance
#  Arguments : Unknown see the description in appropriate subclass
#  Returns   : Mentat::Message::Correlator::Context::Queue::Memory; $self reference
#  Throws    : Dies, if not invoked on object instance or if given invalid arguments
#  Comments  : This method must never be called directly, it is supposed to be overloaded by subclasses.

sub _init
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    ### Initializing Mentat.Message.Correlator.Context.Queue.Memory instance...
    my %config = @_;

    # Call parent version first
    $self = $self->SUPER::_init(%config);

    # Choose what queue type will be used
    my $type = $config{type};
    if ($type) {
        if    ($type eq 'unstable') { $type = "Tie::Array::Sorted"; }
        elsif ($type eq 'stable')   { $type = "Tie::Array::Sorted::Stable"; }
        elsif ($type eq 'lazy')     { $type = "Tie::Array::Sorted::Lazy"; }
        elsif ($type eq 'stazy')    { $type = "Tie::Array::Sorted::Stable::Lazy"; }
    }
    else {
        $type = "Tie::Array::Sorted::Stable";
    }

    my @a;
    tie @a, $type, \&_c;

    # Queue of contexts sorted by timeout
    $self->{QUEUE} = \@a;
    # Map CONTEXT_ID => CONTEXT (allows context fetching by ID)
    $self->{QUEUE_MAP} = {};

    return $self;
}

sub _c { ${$_[0]}->{TIMEOUT} <=> ${$_[1]}->{TIMEOUT} }

=pod

=back

=cut

1;
