package Whois::Mongo;
use strict;
use warnings;

#*******************************************************************************
# DOCUMENTATION SECTION
#*******************************************************************************

=head1 NAME

Whois::Mongo - Simple whois database constructed by parsing whois dump file

=head1 SYNOPSIS

 use Whois::Mongo;

 # Instantinate the whois database
 my $wi_mongo = Whois::Mongo->new('mentat','groups');

 # Find information regarding the given IP address
 my @results = $wi_mongo->find('146.102.1.1');

 # Find information regarding the given IP address
 my $abuse = $wi_mongo->find_abuse('146.102.1.1');

 # Find information regarding the given email address
 my @results = $wi_mongo->rfind('abuse@domain.org');

 # Find information regarding the given email address
 my $networks = $wi_mongo->find_networks('abuse@domain.org');

=head1 DESCRIPTION

Extensive module description

=head1 AUTHOR

 Jan Mach
 Cesnet, z.s.p.o
 jan.mach@cesnet.cz
 http://www.cesnet.cz

=head1 COPYRIGHT

This program is free software; you can redistribute
it and/or modify it under the same terms as Perl itself.

The full text of the license can be found in the
LICENSE file included with this module.

=head1 SEE ALSO

perl(1).

=head1 FUNCTION REFERENCE

=over 4

=cut

#*******************************************************************************
# LIBRARY LOADING SECTION
#*******************************************************************************

#-- Perl core modules ---------------------------------------------------------#
use Carp;
use Scalar::Util 'blessed';

#use Data::Dumper; #-+-> DEVEL ONLY <-+-#
#use Smart::Comments; #-+-> DEVEL ONLY <-+-#

#-- Perl CPAN modules ---------------------------------------------------------#

#-- Custom application modules ------------------------------------------------#
use Whois::Module;
use Mentat::Storage::Mongo;

#*******************************************************************************
# CONSTANTS AND GLOBAL VARIABLES DEFINITION SECTION
#*******************************************************************************

#-- Constants -----------------------------------------------------------------#

#-- Static public class variables (our) ---------------------------------------#

#-- Static protected class variables (my) -------------------------------------#

#*******************************************************************************
# INITIALIZATION AND CLEANUP SECTION
#*******************************************************************************

#-- Module initializations ----------------------------------------------------#
BEGIN {
    use vars qw($VERSION $DEVEL @ISA);
    $VERSION = '0.6';
    $DEVEL   = 0;
    @ISA     = ('Whois::Module');
}


#-- Module clean-up code (global destructor) ----------------------------------#
END {

}

#*******************************************************************************
# IMPLEMENTATION SECTION
#*******************************************************************************

# Realtime mongo query:
## db.groups.find({'networks': { $elemMatch: {'type':'ipv4', 'first': {$lte:3279022145}, 'last': {$gte:3279022145}}}},{_id:1}).pretty()

=item find($) [PUBLIC]

  Usage     : my $result = $instance->find('146.102.1.1');
  Purpose   : Search for given IP address in the inner whois storage
  Arguments : MIXED $ip - IP address [MANDATORY]
  Returns   : ARRAY OF HASH REFERENCES containing network structures
  Throws    : Dies, if not invoked on object instance

=cut

sub find($)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);

    $self->_read_whois_db('update');
    return $self->SUPER::find(@_);
}

=item rfind($) [PUBLIC]

  Usage     : my $result = $instance->rfind('email@abuse.domain');
  Purpose   : Search for given email address in the inner whois storage
  Arguments : MIXED $email - email address [MANDATORY]
  Returns   : ARRAY OF HASH REFERENCES containing network structures
  Throws    : Dies, if not invoked on object instance

=cut

sub rfind($)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);

    $self->_read_whois_db('update');
    return $self->SUPER::rfind(@_);
}

#-------------------------------------------------------------------------------
# Private interface
#-------------------------------------------------------------------------------

# _init($) [PROTECTED]
#
#  Usage     : return $self->_init(@_);
#  Purpose   : Initialize the new Whois::Negistry instance
#  Arguments : STRING $whois_file - name of the file containing whois dump
#              BOOL $force - parse the whois dump file, even if it is too old
#  Returns   : Whois::Negistry reference
#  Throws    : Dies, if not invoked on object instance
#  Comments  : This method must never be called directly, it is supposed to be overloaded by subclasses.

sub _init($$)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($database, $collection, $ttl) = @_;
    die "Whois database name must be given as argument" unless $database;
    die "Whois collection name must be given as argument" unless $collection;

    $self->{DB}  = $database;
    $self->{COL} = $collection;
    $self->{TTL} = $ttl;
    
    $self->{STORAGE} = Mentat::Storage::Mongo->new(database => $self->{DB}, collection => $self->{COL});

    $self->_read_whois_db();

    return $self;
}

# _read_whois_db(;$) [PROTECTED]
#
#  Usage     : $self->_read_whois_db();
#  Purpose   : Populate the inner database by parsing the external whois dump file
#  Arguments : STRING $whois_file - name of the file containing whois dump
#  Returns   : Nothing
#  Throws    : Dies, if not invoked on object instance

sub _read_whois_db(;$)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($update) = @_;

    return if ($update and (not $self->{VALIDITY} or (time() < $self->{VALIDITY})));

    print STDERR "Loading current whois database\n" if $DEVEL;
    $self->_flush();

    my ($groups, $count) = $self->{STORAGE}->find();

    foreach my $g (@$groups) {
        foreach my $r (@{$g->{networks}}) {
            $r->{resolved_abuses} = [$g->{_id}] unless $r->{resolved_abuses};

            $self->_set_network($r->{_id}, $r, 'local-db');
        }
    }

    $self->{VALIDITY} = (time() + $self->{TTL}) if $self->{TTL};
}

=pod

=back

=cut

1;

